"""Class for implementing a Jupyter Cell Dashboard using Python."""

from .containers import StackPanel
from ..serialization import serialize


class Dashboard:
    """The root of a Cell Dashboard."""

    MIMETYPE = "application/vnd.maven.layout+json"

    def __init__(self):
        """Create a new, empty Dashboard."""
        self.__globals = []
        self.__root = StackPanel()

    @property
    def root(self):
        """Get the root stack panel in a dashboard. Read-only."""
        return self.__root

    def add_global(self, name, type_annotation, default=None):
        """Add a global to the dashboard."""
        self.__globals.append({
            "name": name,
            "type": type_annotation,
            "value": serialize(default, type_annotation)
        })

    def __repr__(self):
        """Return a formatted text string with this dashboard's structure."""
        repr_str = "Dashboard " + super().__repr__() + "\n\nGlobals:"
        for gv in self.__globals:
            repr_str += "\n\t" + gv["name"] + " (" + gv["type"] + ")"
        repr_str += "\nRoot " + repr(self.root)
        return repr_str

    def _repr_mimebundle_(self, include=None, exclude=None):
        root, _ = self.root._as_json()
        parts = self.root._collect_parts()
        parts_repr = {}
        for part in parts:
            part_model, part_id = part._as_part()
            parts_repr[part_id] = part_model

        return {
            "text/plain": "Dashboard Layout (if you see this, check your " +
            " plugin install!)",
            self.MIMETYPE: {
                "layout": root,
                "parts": parts_repr,
                "metadata": {},
                "globals": self.__globals,
                "visual": False
            }
        }, {}



"""Server utilities for MavenWorks. Provides the Config handler and Viewer."""

from tornado.web import StaticFileHandler
from notebook.utils import url_path_join
from .local_config import LocalConfig
from .viewer import ViewerApp, ViewerIndex
import os


def load_jupyter_server_extension(nb_app):
    """Register the MavenWorks server extensions."""
    host_pattern = '.*$'
    web_app = nb_app.web_app
    base_url = web_app.settings["base_url"]
    web_app.add_handlers(host_pattern, [
        (
            url_path_join(base_url, r"/serverconfig/?$"),
            LocalConfig,
            {"logger": nb_app.log}
        ), (
            url_path_join(base_url, "/view/", r".*$"),
            ViewerApp
        ), (
            url_path_join(base_url, r"/viewindex/?(.*)$"),
            ViewerIndex
        ), (
            url_path_join(base_url, "/viewer/static", r"(.*)$"),
            StaticFileHandler,
            {
                "path": os.path.abspath(
                    os.path.join(os.path.dirname(__file__), "./static/")
                )
            }
        )
    ])


def _jupyter_server_extension_paths():
    return [{
        "module": "mavenworks.server"
    }]



"""Global settings handler for Jupyter server plugins."""

import os
import json

__all__ = [
    "get_setting",
]

_default_settings = {
    "global_parts_folder": "parts"
}

_local_dir = os.environ.get("CFG_SETTINGS_FILE") or os.path.abspath(
    os.path.dirname(__file__)
)
_settings_file_location = os.path.join(_local_dir, "MavenWorksServerConf.json")
with open(_settings_file_location, "r") as _settings:
    _META_CONFIG = json.load(_settings)


def get_setting(key):
    """Get a particular setting value from the MavenWorksServerConfig."""
    if key in _META_CONFIG:
        return _META_CONFIG[key]
    if key in _default_settings:
        return _default_settings[key]
    return None



"""Module providing a mechanism to use DisplayHandles in Dashboards."""

from IPython.display import DisplayHandle
from typing import Union, Callable, Dict

_new_name_hook: Union[None, Callable] = None
_known_names = {}


def _set_name_hook(cb: Callable):
    global _new_name_hook
    _new_name_hook = cb


def _get_known_names() -> Dict[str, str]:
    return _known_names


def name_display_handle(name: str, handle: DisplayHandle):
    """Mark a DisplayHandle, so that it can be used in a Dashboard.

    Display Handles are simple things you can use to put a cell output into a
    dashboard. If they are named, you can use them in the Visual Designer just
    like any part.

    Display handles are _not_ parts. You won't get overlays, bindings, globals,
    etc. when using them. They are meant for static, or hard-to-integrate,
    visualizations.

    Parameters
    ----------
    name : str
        The name of the region to use. If no name is provided, this display
        handle cannot be used in visual dashboards.
    handle : DisplayHandle
        The handle to name. Once named, the handle can be used in dashboards.

    """
    _known_names[name] = handle.display_id
    if _new_name_hook is not None:
        _new_name_hook(name, handle.display_id)  # pylint: disable=not-callable



"""PartHelpers - MavenWorks internal module.

A set of helpers that support kernel parts. These should not be directly
used by user code, with the exception of :class:PartMetadata.
"""

from collections import OrderedDict
from itertools import islice
from typing import NamedTuple, List
from rx.subjects import Subject


class Option(NamedTuple):
    """Internal data model for options."""

    name: str
    value: any
    type: str


class PartMetadata:
    """A class that describes the metadata about a part.

    This is normally used in :class:KernelPart.get_metadata().
    """

    def __init__(self):
        """Create a new PartMetadata.

        Normally, you would get this by calling super().get_metadata(),
        but if you want to ignore the options defined by a superclass
        and replace them all, then instantiate a new PartMetadata instead.
        """
        # TODO: Format for option metadata
        self.options_bag: List[Option] = []

    def add_option(self, name, default_value=None, type_annotation="Any"):
        self.options_bag.append(Option(name, default_value, type_annotation))

    def __repr__(self):
        return "PartMetadata" + str([
            f'Name={opt.name}, Type={opt.type_annotation}'
            for opt in self.options_bag
        ])


class OptionsBag:
    def __init__(self, metadata: PartMetadata, model=None):
        self.options_bag = OrderedDict()
        self.is_stale = True
        self.OnStale = Subject()
        for opt in metadata.options_bag:
            if model is not None and opt.name in model['options']:
                value = model['options'][opt.name]
            else:
                value = opt.value
            self.options_bag[opt.name] = value
        self.is_stale = False

    def __getitem__(self, item):
        if isinstance(item, int):
            item = next(islice(self.options_bag.keys(), item, item + 1))
        return self.options_bag[item]

    def __setitem__(self, key, value):
        # hack: We allow these sets for now just to keep things working
        if isinstance(key, int):
            key = next(islice(self.options_bag.keys(), key, key + 1))
        self.options_bag[key] = value
        self.set_stale(key, value)

    def __iter__(self):
        return iter(self.options_bag)

    def __del__(self):
        self.OnStale.on_completed()
        del self.options_bag

    def set_stale(self, name, value):
        if self.is_stale:
            return
        self.is_stale = True
        self.OnStale.on_next((name, value))

    def set_fresh(self):
        self.is_stale = False



"""Kernel Part utilities for MavenWorks."""

from .DisplayHandle import name_display_handle
from .interact_wrapper import wrap
from .KernelPart import register_part, KernelPart
from .PartHelpers import Option, OptionsBag
from .PyScatterPart import PyScatterPart
from .WidgetWrapper import gen_wrapper

__all__ = [
    "name_display_handle",
    "wrap",
    "register_part",
    "KernelPart",
    "Option",
    "OptionsBag",
    "PyScatterPart",
    "gen_wrapper"
]



"""Metadata Comm - MavenWorks internal module.

The metadata comm syncs info on metadata about the current MavenWorks install,
such as available kernel parts.

Consumers of this module can update this metadata, so that the client UI can
leverage the updated metadata (for example, third-party parts will appear in
the drag-n-drop dashboard editor).
"""

from ..parts.DisplayHandle import _get_known_names, _set_name_hook
from ..parts.KernelPart import _get_all_parts, _set_new_part_hook
from ..serialization import serialize
from ipykernel.comm import Comm, CommManager
from IPython.core.getipython import get_ipython
from IPython.core.interactiveshell import InteractiveShell


def _send_part(comm: Comm, name, metadata):
    comm.send({
        "msg_type": "new_part",
        "payload": {
            "name": name,
            "options": [
                {
                    "name": opt.name,
                    "type": opt.type,
                    "value": serialize(opt.value, opt.type)
                } for opt in metadata.options_bag
            ]
        }
    })


def _send_new_display_handle(comm: Comm, display_name: str, display_id: str):
    comm.send({
        "msg_type": "named_display_handle",
        "handle_id": display_id,
        "handle_name": display_name
    })


def _send_all_parts(comm: Comm):
    for part_name, part_cls in _get_all_parts():
        _send_part(comm, part_name, part_cls.get_metadata())
    for display_name, display_id in _get_known_names().items():
        _send_new_display_handle(comm, display_name, display_id)
    comm.send({
        "msg_type": "kernel_parts_sent"
    })


def _on_msg(comm: Comm, msg):
    data = msg["content"]["data"]
    msg_type = data["msg_type"]
    if msg_type == "send_parts":
        _send_all_parts(comm)


def _close_comm():
    _set_new_part_hook(None)
    _set_name_hook(None)


def _open_comm(comm: Comm, msg):
    _set_name_hook(
        lambda display_name, display_id: _send_new_display_handle(
            comm,
            display_name,
            display_id
        )
    )
    _set_new_part_hook(
        lambda part_name, part_cls: _send_part(
            comm,
            part_name,
            part_cls.get_metadata()
        )
    )
    comm.on_msg(lambda msg: _on_msg(comm, msg))
    comm.on_close(_close_comm)


ip: InteractiveShell = get_ipython()

if ip is not None:
    manager: CommManager = ip.kernel.comm_manager
    manager.register_target("maven_metadata", _open_comm)



"""A short and sweet part to demonstrate the Python Part API."""
from matplotlib import pyplot as plt
from .KernelPart import KernelPart, register_part
import pandas as pd


@register_part()
class PyScatterPart(KernelPart):
    """A simple KernelPart.

    PyScatterPart takes the following options:

        - Input Data {Table} The data to plot in a scatterplot
        - X Column {String} The column of the above table to use as x-coord
        - Y Column {String} Same, but for the y-coord
    """

    @classmethod
    def get_metadata(cls):
        """Return metadata describing the options this part has."""
        metadata = super().get_metadata()
        df = pd.DataFrame(data={"x": [1, 2, 3], "y": [3, 2, 1]})
        metadata.add_option("Input Data", df, "Table")
        metadata.add_option("X Column", "x", "String")
        metadata.add_option("Y Column", "y", "String")
        return metadata

    def render(self, opts):
        """Render this part.

        :param opts: An OptionsBag containing the data we need to render
        """
        data = opts["Input Data"]
        fig = plt.figure()
        plt.scatter(data[opts["X Column"]], data[opts["Y Column"]])
        plt.close()
        # technically we can also display() it, or just output it using pyplot
        return fig



"""
Serialization utilities
=======================

These are quick and dirty, and lack the flexibility of the serialization system in MavenWorks
All they do is convert between Python and JSON representations of objects.
If you need conversions or pluggable serialization, reimplement this
"""
import json
from itertools import chain
from datetime import date, datetime
from pandas import DataFrame
from numbers import Real
import math
__supports_pyarrow = False
__use_legacy_export = True
try:
    import pyarrow as pa
    [major, minor, _] = pa.__version__.split(".")
    if int(major) == 0 and int(minor) < 15:
        # The JS library doesn't need extra support here
        __use_legacy_export = False
    from io import BytesIO
    __supports_pyarrow = True
except ImportError:
    pass  # no PyArrow support


class PassThrough:
    def __init__(self, data):
        self.data = data

    def to_json(self):
        if isinstance(self.data, str):
            return json.load(self.data)
        return self.data


def _serialize_table(obj):
    if __supports_pyarrow:
        try:
            f = BytesIO()
            data = pa.Table.from_pandas(obj, preserve_index=False)
            if __use_legacy_export:
                batch_writer = pa.RecordBatchFileWriter(f, data.schema, use_legacy_format=True)
            else:
                batch_writer = pa.RecordBatchFileWriter(f, data.schema)
            with batch_writer as writer:
                writer.write_table(data)
            return {
                "arrow": True,
                # Tornado will encode this
                "data": f.getvalue()
            }
        except Exception as e:
            print("Failed to serialize to Arrow")
            print(e)
            # fall through to the JSON format

    def _serialize_row(row_obj, column_types, row_name=None):
        row_data = []
        for i, irow in enumerate(row_obj):
            if (column_types[i] == "Any"):
                datatype = guess_type(irow)
                row_data.append(serialize(irow, datatype))
            else:
                row_data.append(serialize(irow, column_types[i])["value"])
        return {
            "children": [],
            "name": row_name,
            "data": [serialize(irow, column_types[i]) for i, irow in enumerate(row_obj)]
        }

    serialized_table = {
        "rows": [],
        "cols": [],
        "types": []
    }
    for col in obj:
        serialized_table["cols"].append(col)
        # Use the numpy types to infer the MavenType
        serialized_table["types"].append(guess_type(obj.dtypes[col].type, check_instanceof=False))

    for i, row in obj.iterrows():
        serialized_table["rows"].append(_serialize_row(row, serialized_table["types"], i))
    return serialized_table


def _deserialize_table(obj):
    def _deserialize_val(val, idx: int):
        try:
            # if this fails, val doesn't have typeName
            if "typeName" not in val:
                return deserialize({"typeName": obj["types"][idx], "value": val})
            return deserialize(val)
        except TypeError:
            return deserialize({"typeName": obj["types"][idx], "value": val})

    def _deserialize_row(row_obj, nrows):
        row_values = getattr(row_obj, "value", row_obj)  # might be JSON struct
        return chain(
            [getattr(row_values, "name", nrows)],
            [
                _deserialize_val(val, i)
                for i, val in enumerate(row_values["data"])
            ]
        )

    df = DataFrame([_deserialize_row(r, i) for i, r in enumerate(obj["rows"])],
                   columns=list(chain(["rowname"], obj["cols"])))
    return df.set_index("rowname")


def _serialize_number(num_obj):
    if math.isnan(num_obj):
        return None
    return num_obj

def _deserialize_date(date_obj):
    if isinstance(date_obj, str):
        return datetime.fromisoformat(date_obj)
    if isinstance(date_obj, int):
        return datetime.utcfromtimestamp(date_obj/1000)
    raise ValueError('Invalid date object. Type: ' + str(type(date_obj)) + ' Value: ' + str(date_obj))


def _serialize_date(date_obj):
    return date_obj.isoformat()


def serialize(obj, annotated_type):
    """Roughly serialize a Python object into a good approximation of it's MavenWorks-serialized equivalent"""

    if isinstance(obj, PassThrough):
        return obj.to_json()

    # "who needs switch statements anyway?" -Guido Van Rossum
    serializers = {
        "Array": lambda array_obj: list(map(lambda i: serialize(i, guess_type(i)), array_obj)),
        "Date": _serialize_date,
        "DateTime": _serialize_date,
        "Table": _serialize_table,
        "Number": lambda num_obj: _serialize_number(num_obj),
    }
    if obj is not None and annotated_type in serializers:
        return {"typeName": annotated_type, "value": serializers[annotated_type](obj)}
    return {"typeName": annotated_type, "value": obj}


def deserialize(serialized_obj):
    """Attempt a simplistic deserialization of a MavenWorks object into an equivalent Python representation"""
    if serialized_obj is None:
        return None
    try:
        data = serialized_obj["value"]
    except (KeyError, TypeError):
        return serialized_obj  # improperly serialized object, do nothing and hope for the best
    data = serialized_obj["value"]
    deserializers = {
        "Array": lambda obj: list(map(lambda i: deserialize(i), obj)),
        "Table": _deserialize_table,
        "Date": _deserialize_date,
        "DateTime": _deserialize_date,
    }
    if serialized_obj["typeName"] in deserializers:
        return deserializers[serialized_obj["typeName"]](data)
    if hasattr(data, "val") or (isinstance(data, dict) and 'val' in data):
        return data["val"]
    return data


def guess_type(obj, check_instanceof=True):
    """Make a best guess as to what the MavenWorks equivalent type might be"""
    # ordered in decreasing certainty/specificity
    compare = isinstance if check_instanceof else issubclass
    if compare(obj, DataFrame):
        return "Table"
    if compare(obj, BaseException):
        return "Error"
    if compare(obj, datetime):
        return "DateTime"
    if compare(obj, date):
        return "Date"
    if compare(obj, bool):
        return "Boolean"
    if compare(obj, Real):
        # only real-valued numbers are supported in the MavenWorks converter
        return "Number"
    if compare(obj, str):
        return "String"
    if compare(obj, list):
        return "Array"
    if compare(obj, dict):
        return "Object"
    return "Any"  # no support for color, datetime


def traitlet_to_type(trait):
    """Returns an equivalent Maven type annotation for the given traitlet."""
    import traitlets
    if isinstance(trait, traitlets.Bool):
        return "Boolean"
    if any(isinstance(trait, x) for x in [
        traitlets.Integer,
        traitlets.Int,
        traitlets.Long,
        traitlets.Float
    ]):
        return "Number"
    if isinstance(trait, traitlets.Unicode) or isinstance(trait, traitlets.Bytes):
        return "String"
    if any(isinstance(trait, x) for x in [
        traitlets.List,
        traitlets.Tuple,
        traitlets.Set
    ]):
        return "Array"
    if any(isinstance(trait, x) for x in [
        traitlets.Dict,
        traitlets.Instance,
        traitlets.Type,
    ]):
        return "Object"
    if isinstance(trait, traitlets.Union):
        return traitlet_to_type(trait.trait_types[0])
    return "Any"



"""Tornado handler for tracking Maven config (just UDPs for now)."""

import traceback
from notebook.base.handlers import IPythonHandler
from .config_object_tracker import ConfigObjectTracker


class LocalConfig(IPythonHandler):
    """Provides information on config objects via HTTP.

    Requestors can fetch all UDP definitions, and that's about it right now.

    Later passes might add:
      - Read from additional directories
      - Read paths only, and definitions later
      - Read and cache definitions
      - Notify on cache updates
      etc.
    """

    def __init__(self, *args, logger=None, **kwargs):
        """Initialize the HTTP router.

        Takes a logger object as a kwarg for recording errors.
        """
        super().__init__(*args, **kwargs)
        self.logger = logger

    def get(self):
        """Given a working directory, look for any UDPs in the filesystem.

        Parts may live in <Jupyter root>/parts and <workdir>. This means that
        parts can be defined alongside the notebooks that reference them, but
        not above (unless placed into that global /parts folder).
        """
        data = []
        path = self.get_argument("path", ".")
        self.logger.debug("Using path for config request: " + path)
        try:
            iterator = ConfigObjectTracker.get_objects_in_tree('.', path)
            for config_obj in iterator:
                data.append(config_obj)
            self.set_status(200)
            self.write({"objs": data})
        except Exception:
            self.logger.error("Searching config failed")
            self.logger.error(traceback.format_exc())
            self.write_error(500)
        return self.finish()



"""Setup file for MavenWorks python package."""

from setuptools import setup
import os

with open("README.md") as f:
    desc = f.read()

setup(
    name="mavenworks",
    version="0.0.1",
    description="Dashboarding for JupyterLab",
    long_description=desc,
    long_description_content_type="text/markdown",
    author="Mavenomics, Inc",
    license="GPL-3.0 (C) 2019 Mavenomics, Inc.",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Framework :: IPython",
    ],
    packages=[
        "mavenworks",
        "mavenworks.dashboard",
        "mavenworks.parts",
        "mavenworks.services",
        "mavenworks.server",
    ],
    install_requires=[
        "requests",
        "rx",
        "pandas",
        "IPython",
        "ipywidgets",
        "matplotlib",
    ],
    extra_requires={
        "pyarrow-table-interop": [
            "pyarrow"
        ]
    },
    python_requires="~=3.6",
    include_package_data=True
)



"""Interactive dashboard wrapper."""

from .KernelPart import KernelPart, register_part
from ..serialization import guess_type, serialize
from IPython.display import display
from uuid import uuid4

__all__ = [
    "wrap"
]

TYPE_TO_PART = {
    "Number": "SliderPart",
    "String": "TextAreaPart",
    "Date": "DatePickerPart",
    "DateTime": "DatePickerPart",
    "Table": "TableEditor",
    "Boolean": "CheckboxPart"
}

options_for_parts = {
    "SliderPart": ["Value", {
        "Min": serialize(0, "Number"),
        "Max": serialize(500, "Number"),
        "Step": serialize(10, "Number")
    }],
    "TableEditor": ["Input Data", {}],
    "CheckboxPart": ["Checked", {}]
}


def make_part(partType, global_binding_name):
    override, options = options_for_parts[partType] \
        if partType in options_for_parts else ["Value", {}]
    options = options.copy()
    options[override] = {
        "type": "Global",
        "expr": global_binding_name,
        "globals": [global_binding_name]
    }
    return {"application/vnd.maven.part+json": {
        "name": partType,
        "options": options
    }}


def make_wrapper_part(fn, argTypes):
    name = "__autogenerated_" + str(uuid4())
    @register_part(name)  # pylint: disable=unused-variable
    class AutoGeneratedPart(KernelPart):
        @classmethod
        def get_metadata(cls):
            metadata = super().get_metadata()
            for i, argType in enumerate(argTypes):
                metadata.add_option("arg" + str(i), None, argType)

            return metadata

        def render(self, opts):
            vals = [opts[arg] for arg in opts]
            return fn(*vals)

    return name


def wrap(fn, *args):
    """Wrap a function with an interactive MavenWorks dashboard.

    :param fn: A function to wrap, that either returns or ``display()``s some
    output.

    :param args: The initial values to pass to the function. Each argument
    provided here will be bound to an input part.

    """
    parts = {}
    global_defs = []
    arg_types = []
    layout = {
        "properties": {},
        "typeName": 0,
        "uuid": "root",
        "children": []
    }
    for i, arg in enumerate(args):
        arg_type = guess_type(arg)
        if arg_type not in TYPE_TO_PART:
            pass
        arg_types.append(arg_type)
        global_defs.append({
            "name": "arg" + str(i),
            "type": arg_type,
            "value": serialize(arg, arg_type)["value"]
        })
        part_id = str(uuid4())
        parts[part_id] = make_part(TYPE_TO_PART[arg_type], "arg" + str(i))
        layout["children"].append({
            "properties": {
                "caption": "Arg" + str(i)
            },
            "typeName": 1,
            "uuid": str(uuid4()),
            "guid": part_id
        })
    new_part = make_wrapper_part(fn, arg_types)
    part_id = str(uuid4())
    parts[part_id] = {"application/vnd.maven.part+json": {
        "name": new_part,
        "options": {"arg" + str(i): {
            "type": "Global",
            "expr": "arg" + str(i),
            "globals": ["arg" + str(i)]
        } for i in range(len(arg_types))}
    }}
    layout["children"].append({
        "properties": {
            "caption": "Output"
        },
        "typeName": 1,
        "uuid": str(uuid4()),
        "guid": part_id
    })
    display({"application/vnd.maven.layout+json": {
        "layout": layout,
        "globals": global_defs,
        "parts": parts
    }}, raw=True)



"""Private utilities for translating kwargs to proper layout props."""

__attached_properties__ = [
    "Fixed Size (px)",
    "Stretch",
    "Canvas Panel.Rectangle",
    "Grid Panel.Column",
    "Grid Panel.Column Span",
    "Grid Panel.Row",
    "Grid Panel.Row Span",
]

__kwarg_to_property_map__ = {
    # kwarg name                real name
    "border_width":             "borderWidth.px",
    "border_color":             "borderColor",
    "background_color":         "backgroundColor",
    "padding":                  "padding.px",
    "show_title":               "showTitle",
    "show_region":              "showRegion",
    "caption":                  "caption",
    "caption_color":            "captionColor",
    "caption_background":       "captionBackground",

    # WidgetLayoutRegion
    "show_overlays":            "showOverlays",

    # StackPanel
    "horizontal":               "horizontal",
    "show_splitters":           "showSplitters",
    "splitter_width":           "splitterWidth.px",

    # Grid
    "n_rows":                   "nRows",
    "n_cols":                   "nCols",
    "grid_spacing":             "spacing.px",

    # TabPanel
    "foreground_index":         "ForegroundIndex",

    # Attached properties
    "Stack_FixedSize":          "Fixed Size (px)",
    "Stack_Stretch":            "Stretch",
    "Canvas_Rectangle":         "Canvas Panel.Rectangle",
    "Grid_Column":              "Grid Panel.Column",
    "Grid_ColSpan":             "Grid Panel.Column Span",
    "Grid_Row":                 "Grid Panel.Row",
    "Grid_RowSpan":             "Grid Panel.Row Span",
}


def kwargs_to_properties(kwargs):
    """Translate the given kwargs to a set of attached and layout props.

    The layout props can be directly inserted to the layout region's properties
    in the JSON model. The attached properties must be passed up to the region
    parent.
    """
    _attached_props = {}
    _layout_props = {}

    for key, value in kwargs.items():
        real_key = __kwarg_to_property_map__[key]
        if real_key in __attached_properties__:
            _attached_props[real_key] = value
        else:
            _layout_props[real_key] = value

    return _attached_props, _layout_props



class Binding:
    def __init__(self, binding_type, expr):
        self.__type = binding_type
        self.__expr = expr

    def _as_json(self):
        return {
            "type": self.__type,
            "expr": self.__expr,
            "__detect_globals": True
        }


class Mql(Binding):
    def __init__(self, query):
        super().__init__("Mql", query)


class Js(Binding):
    def __init__(self, code):
        super().__init__("JavaScript", code)


class Py(Binding):
    def __init__(self, code):
        super().__init__("Eval", code)


class Global(Binding):
    def __init__(self, global_name):
        super().__init__("Global", global_name)
